import argparse
import os
import random
import shutil
import time
import warnings
from enum import Enum

import torch
seed = 1337
torch.manual_seed(seed)
torch.cuda.manual_seed(seed)
torch.backends.cudnn.deterministic=True

import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.multiprocessing as mp
import torch.nn as nn
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import torchvision.datasets as datasets
import torchvision.models as models
import torchvision.transforms as transforms
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import Subset
from resnet import resnet50, resnext50_32x4d, wide_resnet50_2
import torch.nn.functional as F
import torchvision

import numpy as np
import random
# random.seed(seed)
def _init_fn(worker_id):
    np.random.seed(int(42))

from loss import Discriminators_1l, Discriminators_2l, VariationalClassification, LGM
from evaluation import Evaluator

model_names = sorted(name for name in models.__dict__
    if name.islower() and not name.startswith("__")
    and callable(models.__dict__[name]))

parser = argparse.ArgumentParser(description='PyTorch ImageNet Training')
parser.add_argument('data', metavar='DIR', nargs='?', default='imagenet',
                    help='path to dataset (default: imagenet)')
parser.add_argument('-a', '--arch', metavar='ARCH', default='resnet18',
                    choices=model_names,
                    help='model architecture: ' +
                        ' | '.join(model_names) +
                        ' (default: resnet18)')
parser.add_argument('-j', '--workers', default=1, type=int, metavar='N',
                    help='number of data loading workers (default: 4)')
parser.add_argument('--epochs', default=90, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N',
                    help='mini-batch size (default: 256), this is the total '
                         'batch size of all GPUs on the current node when '
                         'using Data Parallel or Distributed Data Parallel')
parser.add_argument('--lr', '--learning-rate', default=0.1, type=float,
                    metavar='LR', help='initial learning rate', dest='lr')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--wd', '--weight-decay', default=1e-3, type=float,
                    metavar='W', help='weight decay (default: 1e-4)',
                    dest='weight_decay')
parser.add_argument('-p', '--print-freq', default=10, type=int,
                    metavar='N', help='print frequency (default: 10)')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
parser.add_argument('--pretrained', dest='pretrained', action='store_true',
                    help='use pre-trained model')
parser.add_argument('--world-size', default=-1, type=int,
                    help='number of nodes for distributed training')
parser.add_argument('--rank', default=-1, type=int,
                    help='node rank for distributed training')
parser.add_argument('--dist-url', default='tcp://224.66.41.62:23456', type=str,
                    help='url used to set up distributed training')
parser.add_argument('--dist-backend', default='nccl', type=str,
                    help='distributed backend')
parser.add_argument('--seed', default=None, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--gpu', default=None, type=int,
                    help='GPU id to use.')
parser.add_argument('--multiprocessing-distributed', action='store_true',
                    help='Use multi-processing distributed training to launch '
                         'N processes per node, which has N GPUs. This is the '
                         'fastest way to use PyTorch for either single node or '
                         'multi node data parallel training')
parser.add_argument('--dummy', action='store_true', help="use fake data to benchmark")
parser.add_argument('--vc', action='store_true', help="use variational classification")
parser.add_argument('--lgm', action='store_true', help="use variational classification")

parser.add_argument('--online_log', '--ol', action='store_true', help="=online_log")
parser.add_argument('--dont_save', '--ds', action='store_true', help="=online_log")

parser.add_argument('--disc_layers', default=1, type=int,
                    help='layers in discriminator')
parser.add_argument('--l1', default=0.1, type=float,
                    help='layers in discriminator')
parser.add_argument('--l2', default=0.01, type=float,
                    help='layers in discriminator')
parser.add_argument('--num_classes', default=1000, type=int,
                    help='num classes')

parser.add_argument('--desc', default="", type=str,
                    help='num classes')


parser.add_argument('--run_name', default='00', type=str,
                    help='dataset ')



parser.add_argument('--dataset', default='tiny', type=str,
                    help='dataset ')


best_acc1 = 0
class full_model(nn.Module):
    def __init__(self, args, network, num_classes=1000, num_feats=512 ) -> None:
        super(full_model, self).__init__()



        self.network = network
        if args.lgm:
            self.criterion = LGM(args, num_classes, num_feats)

        if args.vc:
            self.criterion = VariationalClassification(args, num_classes, num_feats)
    def forward(self, inputs):
        return self.network(inputs)

        # tsampled_accuracy += ((Tsampled < 0) == torch.ones_like(Tsampled)).sum().item()
        # treal_accuracy += ((Treal > 0) == torch.ones_like(Treal)).sum().item()

        


def main():
    args = parser.parse_args()
    print(args)



    args.loss = "CE"
    if args.vc:
        args.loss = "VC"
    elif args.lgm:
        args.loss = "LGM"

    
    import uuid
    if args.run_name == "00":
            args.run_name = uuid.uuid4().__str__()[:5] +'_'+ args.loss + '_' + args.desc
    print("The run name is {}".format(args.run_name))


    


    if args.online_log:
        pass


    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        cudnn.deterministic = True
        cudnn.benchmark = False
        warnings.warn('You have chosen to seed training. '
                      'This will turn on the CUDNN deterministic setting, '
                      'which can slow down your training considerably! '
                      'You may see unexpected behavior when restarting '
                      'from checkpoints.')

    if args.gpu is not None:
        warnings.warn('You have chosen a specific GPU. This will completely '
                      'disable data parallelism.')

    if args.dist_url == "env://" and args.world_size == -1:
        args.world_size = int(os.environ["WORLD_SIZE"])

    args.distributed = args.world_size > 1 or args.multiprocessing_distributed

    if torch.cuda.is_available():
        ngpus_per_node = torch.cuda.device_count()
    else:
        ngpus_per_node = 1
    if args.multiprocessing_distributed:
        # Since we have ngpus_per_node processes per node, the total world_size
        # needs to be adjusted accordingly
        args.world_size = ngpus_per_node * args.world_size
        # Use torch.multiprocessing.spawn to launch distributed processes: the
        # main_worker process function
        mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, args))
    else:
        # Simply call main_worker function
        main_worker(args.gpu, ngpus_per_node, args)


def main_worker(gpu, ngpus_per_node, args):
    global best_acc1
    global best_ece
    args.gpu = gpu

    if args.dataset == "imagenet" :
        args.num_classes = 1000
    elif args.dataset == "tiny":
        args.num_classes = 200
    elif args.dataset == "camelyon":
        args.num_classes  = 2

    if args.gpu is not None:
        print("Use GPU: {} for training".format(args.gpu))

    if args.distributed:
        if args.dist_url == "env://" and args.rank == -1:
            args.rank = int(os.environ["RANK"])
        if args.multiprocessing_distributed:
            # For multiprocessing distributed training, rank needs to be the
            # global rank among all the processes
            args.rank = args.rank * ngpus_per_node + gpu
        dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                world_size=args.world_size, rank=args.rank)
    # create model

    if args.pretrained:
        print("=> using pre-trained model '{}'".format(args.arch))
        model = models.__dict__[args.arch](pretrained=True)
    else:
        print("=> creating model '{}'".format(args.arch))

        model = full_model(args, network=resnet50(num_classes = args.num_classes), num_classes=args.num_classes, num_feats=2048 )
        if args.dataset == "tiny":
            model.network.conv1 = nn.Conv2d(3,64, kernel_size=(3,3), stride=(2,2), padding=(1,1), bias=False)
            model.network.maxpool = nn.Sequential()
            model.network.avgpool = nn.AdaptiveAvgPool2d(1)




    if not torch.cuda.is_available() and not torch.backends.mps.is_available():
        print('using CPU, this will be slow')
    elif args.distributed:
        # For multiprocessing distributed, DistributedDataParallel constructor
        # should always set the single device scope, otherwise,
        # DistributedDataParallel will use all available devices.
        if torch.cuda.is_available():
            if args.gpu is not None:
                torch.cuda.set_device(args.gpu)
                model.cuda(args.gpu)
                # When using a single GPU per process and per
                # DistributedDataParallel, we need to divide the batch size
                # ourselves based on the total number of GPUs of the current node.
                args.batch_size = int(args.batch_size / ngpus_per_node)
                args.workers = int((args.workers + ngpus_per_node - 1) / ngpus_per_node)
                model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
            else:
                model.cuda()
                # DistributedDataParallel will divide and allocate batch_size to all
                # available GPUs if device_ids are not set
                model = torch.nn.parallel.DistributedDataParallel(model)
    elif args.gpu is not None and torch.cuda.is_available():
        torch.cuda.set_device(args.gpu)
        model = model.cuda(args.gpu)
    #elif torch.backends.mps.is_available():
    #    device = torch.device("mps")
    #    model = model.to(device)
    else:
        # DataParallel will divide and allocate batch_size to all available GPUs
        if args.arch.startswith('alexnet') or args.arch.startswith('vgg'):
            model.features = torch.nn.DataParallel(model.features)
            model.cuda()
        else:
            model = torch.nn.DataParallel(model).cuda()

    if torch.cuda.is_available():
        if args.gpu:
            device = torch.device('cuda:{}'.format(args.gpu))
        else:
            device = torch.device("cuda")
    elif torch.backends.mps.is_available():
        device = torch.device("mps")
    else:
        device = torch.device("cpu")
    # define loss function (criterion), optimizer, and learning rate scheduler
    if args.vc or args.lgm:
        criterion = model.module.criterion

        optimizer_2 = torch.optim.SGD([p for p in model.module.criterion.VC.parameters()], args.lr*0.1,
                                momentum=args.momentum,
                                weight_decay=args.weight_decay, nesterov=True)

        # optimizer_2 = torch.optim.AdamW([p for p in model.module.criterion.VC.parameters()])

        # optimizer_2 = torch.optim.Adam([p for p in model.module.criterion.VC.parameters()], 1e-3,
        #                     weight_decay=args.weight_decay)
    else: 
        optimizer_2 = None
        criterion = nn.CrossEntropyLoss() 


    optimizer_1 = torch.optim.SGD([p for p in model.module.network.parameters()] , args.lr,
                                momentum=args.momentum,
                                weight_decay=args.weight_decay, nesterov=True)

    # optimizer_1 = torch.optim.AdamW([p for p in model.module.network.parameters()] , )

                    
    
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    scheduler_1 = StepLR(optimizer_1, step_size=30, gamma=0.1)
    if args.vc or args.lgm:
        scheduler_2 = StepLR(optimizer_2, step_size=30, gamma=0.1)
    
    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            if args.gpu is None:
                checkpoint = torch.load(args.resume)
            elif torch.cuda.is_available():
                # Map model to be loaded to specified single gpu.
                loc = 'cuda:{}'.format(args.gpu)
                checkpoint = torch.load(args.resume, map_location=loc)
            args.start_epoch = checkpoint['epoch']
            best_acc1 = checkpoint['best_acc1']
            if args.gpu is not None:
                # best_acc1 may be from a checkpoint from a different GPU
                best_acc1 = best_acc1.to(args.gpu)
            model.load_state_dict(checkpoint['state_dict'])
            optimizer_1.load_state_dict(checkpoint['optimizer_1'])
            scheduler_1.load_state_dict(checkpoint['scheduler_1'])

            if args.vc or args.lgm:

                optimizer_2.load_state_dict(checkpoint['optimizer_2'])
                scheduler_2.load_state_dict(checkpoint['scheduler_2'])
            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.resume, checkpoint['epoch']))
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))


    # Data loading code
    if args.dummy:
        print("=> Dummy data is used!")
        train_dataset = datasets.FakeData(1281167, (3, 224, 224), 1000, transforms.ToTensor())
        val_dataset = datasets.FakeData(50000, (3, 224, 224), 1000, transforms.ToTensor())
    elif args.dataset in ["tiny", "imagenet"]:
        traindir = os.path.join(args.data, 'train')
        valdir = os.path.join(args.data, 'val')
        
        if args.dataset == "imagenet":
            normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                        std=[0.229, 0.224, 0.225])

            train_dataset = datasets.ImageFolder(
                traindir,
                transforms.Compose([
                    transforms.RandomResizedCrop(224),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    normalize,
                ]))

            val_dataset = datasets.ImageFolder(
                valdir,
                transforms.Compose([
                    transforms.Resize(256),
                    transforms.CenterCrop(224),
                    transforms.ToTensor(),
                    normalize,
                ]))
        if args.dataset == "tiny":

            train_dataset = datasets.ImageFolder(
                    traindir,
                    transforms.Compose([
                transforms.RandomRotation(20),
                transforms.RandomHorizontalFlip(0.5),
                transforms.ToTensor(),
                transforms.Normalize([0.4802, 0.4481, 0.3975], [0.2302, 0.2265, 0.2262]),
                ]))

            val_dataset = datasets.ImageFolder(
            valdir,
            transforms.Compose(

                [
                transforms.ToTensor(),
                transforms.Normalize([0.4802, 0.4481, 0.3975], [0.2302, 0.2265, 0.2262]),
            ]
            ))

        
        if args.distributed:
            train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
            val_sampler = torch.utils.data.distributed.DistributedSampler(val_dataset, shuffle=False, drop_last=False)
        else:
            train_sampler = None
            val_sampler = None

        train_loader = torch.utils.data.DataLoader(
            train_dataset, batch_size=args.batch_size, shuffle=(train_sampler is None),
            num_workers=args.workers, pin_memory=True, sampler=train_sampler)

        val_loader = torch.utils.data.DataLoader(
            val_dataset, batch_size=args.batch_size, shuffle=False,
            num_workers=args.workers, pin_memory=True, sampler=val_sampler)
    else:

        mean_ = {
            'cifar10': (0.4914, 0.4822, 0.4465),
            'cifar100': (0.5071, 0.4867, 0.4408),
            'mnist' : (0.1307),
            'imagenet': [0.485, 0.456, 0.406],
            'svhn': (0.4914, 0.4822, 0.4465)
        }

        std_ = {
            'cifar10': (0.2023, 0.1994, 0.2010),
            'cifar100': (0.2675, 0.2565, 0.2761),
            'mnist': (0.3081,),
            'imagenet': [0.229, 0.224, 0.225],
            'svhn': (0.2023, 0.1994, 0.2010)
        }
        if args.dataset in std_.keys():
            transform_train = transforms.Compose([
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize(mean_[args.dataset], std_[args.dataset]),
            ]) # meanstd transformation

            transform_test = transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize(mean_[args.dataset], std_[args.dataset]),
            ])



        # traindir = os.path.join("/hdd/imnet", 'train')
        # valdir = os.path.join("/hdd/imnet", 'val')
        # normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
        #                              std=[0.229, 0.224, 0.225])

       
        # if(args.dataset == 'mnist'):

        #     train_dataset = torchvision.datasets.MNIST(root='./data', train=True, download=True, transform=transform_test)
        #     test_dataset = torchvision.datasets.MNIST(root='./data', train=False, download=False, transform=transform_test)
        #     args.num_classes = 10
        # elif(args.dataset == 'cifar10'):

        #     train_dataset = torchvision.datasets.CIFAR10(root='./data', train=True, download=True, transform=transform_train)
        #     train_dataset, val_dataset = torch.utils.data.random_split(train_dataset, [int(0.9*len(train_dataset)), int(0.1*len(train_dataset))])
        #     test_dataset = torchvision.datasets.CIFAR10(root='./data', train=False, download=False, transform=transform_test)
        #     args.num_classes = 10
        # elif(args.dataset == 'cifar100'):

        #     train_dataset = torchvision.datasets.CIFAR100(root='./data', train=True, download=True, transform=transform_train)
        #     train_dataset, val_dataset = torch.utils.data.random_split(train_dataset, [int(0.9*len(train_dataset)), int(0.1*len(train_dataset))])
        #     test_dataset = torchvision.datasets.CIFAR100(root='./data', train=False, download=False, transform=transform_test)
        #     args.num_classes = 100





        # train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True, num_workers=2, worker_init_fn=_init_fn)
        # val_loader = torch.utils.data.DataLoader(val_dataset, batch_size=args.batch_size, shuffle=False, num_workers=2, worker_init_fn=_init_fn)
        # test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=args.batch_size, shuffle=False, num_workers=2, worker_init_fn=_init_fn)






        # train_dataset = datasets.ImageFolder(
        #     traindir,
        #     transforms.Compose([
        #         transforms.RandomResizedCrop(224),
        #         transforms.RandomHorizontalFlip(),
        #         transforms.ToTensor(),
        #         normalize,
        #     ]))

        # val_dataset = datasets.ImageFolder(
        #     valdir,
        #     transforms.Compose([
        #         transforms.Resize(256),
        #         transforms.CenterCrop(224),
        #         transforms.ToTensor(),
        #         normalize,
        #     ]))

        # train_dataset = datasets.ImageFolder(
        #     valdir,
        #     transforms.Compose([
        #         transforms.Resize(256),
        #         transforms.CenterCrop(224),
        #         transforms.ToTensor(),
        #         normalize,
        #     ]))

    params = {}
    if args.evaluate:
        acc1, ECE =  validate(val_loader, model, criterion, args)
        print("Acc: {}".format(acc1))
        print("ECE: {}".format(ECE))
        return acc1, ECE

    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            train_sampler.set_epoch(epoch)

        # train for one epoch
        train(train_loader, model, criterion, [optimizer_1, optimizer_2], epoch, device, args)

        # evaluate on validation set
        acc1, ECE = validate(val_loader, model, criterion, args)
        
        scheduler_1.step()
        # if args.vc or args.lgm:
        #     scheduler_2.step()
        
        # remember best acc@1 and save checkpoint
        is_best = acc1 > best_acc1
        if is_best:
            best_ece = ECE
        best_acc1 = max(acc1, best_acc1)
        params["best_acc"] = best_acc1
        params["acc"] = acc1
        params["ECE"] = ECE
        params["best_e"] = best_ece


        if args.online_log:
            import wandb
            wandb.log(params)

        
        print("best accuracy: {}".format(best_acc1))
        print("best ECE: {}".format(best_ece))

        if not args.dont_save:
            if not args.multiprocessing_distributed or (args.multiprocessing_distributed
                    and args.rank % ngpus_per_node == 0):
                if args.vc or args.lgm:


                    save_checkpoint({
                        'epoch': epoch + 1,
                        'args': args,
                        'arch': args.arch,
                        'state_dict': model.state_dict(),
                        'best_acc1': best_acc1,
                        'best_ece': best_ece,
                        'optimizer_1' : optimizer_1.state_dict(),
                        'optimizer_2' : optimizer_2.state_dict(),
                        'scheduler_1' : scheduler_1.state_dict(),
                        'scheduler_2' : scheduler_2.state_dict(),
                    }, is_best, filename="{}_checkpoint.tar".format(args.run_name))
                
                else:
                    save_checkpoint({
                        'epoch': epoch + 1,
                        'args': args,
                        'arch': args.arch,
                        'state_dict': model.state_dict(),
                        'best_acc1': best_acc1,
                        'best_ece': best_ece,
                        'optimizer_1' : optimizer_1.state_dict(),
                        'scheduler_1' : scheduler_1.state_dict(),
                    }, is_best, filename="{}_checkpoint.tar".format(args.run_name))


def train(train_loader, model, criterion, optimizer, epoch, device, args):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        len(train_loader),
        [batch_time, data_time, losses, top1, top5],
        prefix="Epoch: [{}]".format(epoch))

    # switch to train mode
    model.train()
    optimizer_1, optimizer_2 = optimizer

    end = time.time()
    for i, bb in enumerate(train_loader):

        # measure data loading time
        data_time.update(time.time() - end)

        # move data to the same device as model
        images = images.to(device, non_blocking=True)
        target = target.to(device, non_blocking=True)

        output, logits = model(images)

        if args.vc or args.lgm:
            loss, logits = criterion(output, target)
        else:
            loss = criterion(logits, target)

        # measure accuracy and record loss
        if args.dataset == "camelyon":
            acc1, acc5 = accuracy(logits, target, topk=(1, 1))
        else:

            acc1, acc5 = accuracy(logits, target, topk=(1, 5))
        losses.update(loss.item(), images.size(0))
        top1.update(acc1[0], images.size(0))
        top5.update(acc5[0], images.size(0))

        # compute gradient and do SGD step
        optimizer_1.zero_grad()
        if args.vc or args.lgm:
            optimizer_2.zero_grad()
        loss.backward()
        optimizer_1.step()
        if args.vc or args.lgm:
            optimizer_2.step()

        if args.vc or args.lgm:
            optimizer_1.zero_grad()
            optimizer_2.zero_grad()
            if args.vc:
                criterion.discriminator_train()
            optimizer_1.zero_grad()
            optimizer_2.zero_grad()
        

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            progress.display(i + 1)
        


def validate(val_loader, model, criterion, args):

    def run_validate(loader, base_progress=0):

        with torch.no_grad():
            end = time.time()
            
            for i, bb in enumerate(loader):

                i = base_progress + i
                if args.gpu is not None and torch.cuda.is_available():
                    images = images.cuda(args.gpu, non_blocking=True)
                #if torch.backends.mps.is_available():
                #    images = images.to('mps')
                #    target = target.to('mps')
                if torch.cuda.is_available():
                    target = target.cuda(args.gpu, non_blocking=True)

                # compute output
                output, logits = model(images)
                if args.vc or args.lgm:
                    loss, logits = criterion(output, target)
                else:
                    loss = criterion(output, target)

                # measure accuracy and record loss

                _, predicted = torch.max(logits.data, 1)

                accuracies.extend(predicted.eq(target.data).cpu().numpy().tolist())
                confidences.extend(F.softmax(logits, dim=-1)[torch.arange(target.shape[0]), predicted].cpu().numpy().tolist())



                acc1, acc5 = accuracy(logits, target, topk=(1, 5))
                losses.update(loss.item(), images.size(0))
                top1.update(acc1[0], images.size(0))
                top5.update(acc5[0], images.size(0))

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                if i % args.print_freq == 0:
                    progress.display(i + 1)

    
    
    batch_time = AverageMeter('Time', ':6.3f', Summary.NONE)
    losses = AverageMeter('Loss', ':.4e', Summary.NONE)
    top1 = AverageMeter('Acc@1', ':6.2f', Summary.AVERAGE)
    top5 = AverageMeter('Acc@5', ':6.2f', Summary.AVERAGE)
    progress = ProgressMeter(
        len(val_loader) + (args.distributed and (len(val_loader.sampler) * args.world_size < len(val_loader.dataset))),
        [batch_time, losses, top1, top5],
        prefix='Test: ')

    # switch to evaluate mode



    accuracies = []
    confidences = []
    model.eval()




    run_validate(val_loader)


    ECE, MCE, _, _, _ = Evaluator.create_fine_bins(accuracies, confidences)
    ECE = ECE * 100
    if args.distributed:
        top1.all_reduce()
        top5.all_reduce()

    if args.distributed and (len(val_loader.sampler) * args.world_size < len(val_loader.dataset)):
        aux_val_dataset = Subset(val_loader.dataset,
                                 range(len(val_loader.sampler) * args.world_size, len(val_loader.dataset)))
        aux_val_loader = torch.utils.data.DataLoader(
            aux_val_dataset, batch_size=args.batch_size, shuffle=False,
            num_workers=args.workers, pin_memory=True)
        run_validate(aux_val_loader, len(val_loader))

    progress.display_summary()

    return top1.avg, ECE


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
    filename = "models/" + filename
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, '{}_best'.format(filename))

class Summary(Enum):
    NONE = 0
    AVERAGE = 1
    SUM = 2
    COUNT = 3

class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self, name, fmt=':f', summary_type=Summary.AVERAGE):
        self.name = name
        self.fmt = fmt
        self.summary_type = summary_type
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def all_reduce(self):
        if torch.cuda.is_available():
            device = torch.device("cuda")
        elif torch.backends.mps.is_available():
            device = torch.device("mps")
        else:
            device = torch.device("cpu")
        total = torch.tensor([self.sum, self.count], dtype=torch.float32, device=device)
        dist.all_reduce(total, dist.ReduceOp.SUM, async_op=False)
        self.sum, self.count = total.tolist()
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)
    
    def summary(self):
        fmtstr = ''
        if self.summary_type is Summary.NONE:
            fmtstr = ''
        elif self.summary_type is Summary.AVERAGE:
            fmtstr = '{name} {avg:.3f}'
        elif self.summary_type is Summary.SUM:
            fmtstr = '{name} {sum:.3f}'
        elif self.summary_type is Summary.COUNT:
            fmtstr = '{name} {count:.3f}'
        else:
            raise ValueError('invalid summary type %r' % self.summary_type)
        
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print('\t'.join(entries))
        
    def display_summary(self):
        entries = [" *"]
        entries += [meter.summary() for meter in self.meters]
        print(' '.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'

def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


if __name__ == '__main__':
    main()
